@@ -59,9 +59,11 @@
  *
  */
 
-#include <config.h>
+#include "engine_pkcs11.h"
 #include <stdio.h>
 #include <string.h>
+#include <openssl/opensslv.h>
+#include <openssl/opensslconf.h>
 #include <openssl/crypto.h>
 #include <openssl/objects.h>
 #include <openssl/engine.h>
@@ -70,7 +72,6 @@
 #error did not get engine.h
 #endif
 #include <libp11.h>
-#include "engine_pkcs11.h"
 
 #define PKCS11_ENGINE_ID "pkcs11"
 #define PKCS11_ENGINE_NAME "pkcs11 engine"
@@ -85,7 +86,7 @@
 
 static int pkcs11_engine_destroy(ENGINE * e);
 static int pkcs11_engine_ctrl(ENGINE * e, int cmd, long i, void *p,
-			      void (*f) ());
+	void (*f) ());
 
 /* The definitions for control commands specific to this engine */
 
@@ -93,45 +94,48 @@ static int pkcs11_engine_ctrl(ENGINE * e, int cmd, long i, void *p,
 
 static const ENGINE_CMD_DEFN pkcs11_cmd_defns[] = {
 	{CMD_SO_PATH,
-	 "SO_PATH",
-	 "Specifies the path to the 'pkcs11-engine' shared library",
-	 ENGINE_CMD_FLAG_STRING},
+		"SO_PATH",
+		"Specifies the path to the 'pkcs11-engine' shared library",
+		ENGINE_CMD_FLAG_STRING},
 	{CMD_MODULE_PATH,
-	 "MODULE_PATH",
-	 "Specifies the path to the pkcs11 module shared library",
-	 ENGINE_CMD_FLAG_STRING},
+		"MODULE_PATH",
+		"Specifies the path to the pkcs11 module shared library",
+		ENGINE_CMD_FLAG_STRING},
 	{CMD_PIN,
-	 "PIN",
-	 "Specifies the pin code",
-	 ENGINE_CMD_FLAG_STRING},
+		"PIN",
+		"Specifies the pin code",
+		ENGINE_CMD_FLAG_STRING},
 	{CMD_VERBOSE,
-	 "VERBOSE",
-	 "Print additional details",
-	 ENGINE_CMD_FLAG_NO_INPUT},
+		"VERBOSE",
+		"Print additional details",
+		ENGINE_CMD_FLAG_NO_INPUT},
 	{CMD_QUIET,
-	 "QUIET",
-	 "Remove additional details",
-	 ENGINE_CMD_FLAG_NO_INPUT},
+		"QUIET",
+		"Remove additional details",
+		ENGINE_CMD_FLAG_NO_INPUT},
 	{CMD_LOAD_CERT_CTRL,
-	 "LOAD_CERT_CTRL",
-	 "Get the certificate from card",
-	 ENGINE_CMD_FLAG_INTERNAL},
+		"LOAD_CERT_CTRL",
+		"Get the certificate from card",
+		ENGINE_CMD_FLAG_INTERNAL},
 	{CMD_INIT_ARGS,
-	 "INIT_ARGS",
-	 "Specifies additional initialization arguments to the pkcs11 module",
-	 ENGINE_CMD_FLAG_STRING},
+		"INIT_ARGS",
+		"Specifies additional initialization arguments to the pkcs11 module",
+		ENGINE_CMD_FLAG_STRING},
 	{0, NULL, NULL, 0}
 };
 
 /* Destructor */
 static int pkcs11_engine_destroy(ENGINE * e)
 {
+	(void)e;
 	return 1;
 }
 
 static int pkcs11_engine_ctrl(ENGINE * e, int cmd, long i, void *p,
-			      void (*f) ())
+		void (*f) ())
 {
+	(void)i; /* We don't currently take integer parameters */
+	(void)f; /* We don't currently take callback parameters */
 	/*int initialised = ((pkcs11_dso == NULL) ? 0 : 1); */
 	switch (cmd) {
 	case CMD_MODULE_PATH:
@@ -150,50 +154,30 @@ static int pkcs11_engine_ctrl(ENGINE * e, int cmd, long i, void *p,
 	return 0;
 }
 
-#if 0
-/* set up default rsa_meth_st with overloaded rsa functions */
-/* the actual implementation needs to be in another object */
-
-static int (*orig_finish) (RSA * rsa);
-
-static int pkcs11_engine_rsa_finish(RSA * rsa)
-{
-
-	pkcs11_rsa_finish(rsa);
-
-	if (orig_finish)
-		orig_finish(rsa);
-	return 1;
-
-}
-#endif
-
 /* This internal function is used by ENGINE_pkcs11() and possibly by the
  * "dynamic" ENGINE support too */
 static int bind_helper(ENGINE * e)
 {
 	if (!ENGINE_set_id(e, PKCS11_ENGINE_ID) ||
-	    !ENGINE_set_destroy_function(e, pkcs11_engine_destroy) ||
-	    !ENGINE_set_init_function(e, pkcs11_init) ||
-	    !ENGINE_set_finish_function(e, pkcs11_finish) ||
-	    !ENGINE_set_ctrl_function(e, pkcs11_engine_ctrl) ||
-	    !ENGINE_set_cmd_defns(e, pkcs11_cmd_defns) ||
-	    !ENGINE_set_name(e, PKCS11_ENGINE_NAME) ||
+			!ENGINE_set_destroy_function(e, pkcs11_engine_destroy) ||
+			!ENGINE_set_init_function(e, pkcs11_init) ||
+			!ENGINE_set_finish_function(e, pkcs11_finish) ||
+			!ENGINE_set_ctrl_function(e, pkcs11_engine_ctrl) ||
+			!ENGINE_set_cmd_defns(e, pkcs11_cmd_defns) ||
+			!ENGINE_set_name(e, PKCS11_ENGINE_NAME) ||
 #ifndef OPENSSL_NO_RSA
-	    !ENGINE_set_RSA(e, PKCS11_get_rsa_method()) ||
+			!ENGINE_set_RSA(e, PKCS11_get_rsa_method()) ||
 #endif
-#ifndef OPENSSL_NO_DSA
-	    !ENGINE_set_DSA(e, DSA_get_default_method()) ||
+#ifndef OPENSSL_NO_EC
+#ifndef OPENSSL_NO_ECDSA
+			!ENGINE_set_ECDSA(e, PKCS11_get_ecdsa_method()) ||
 #endif
-#ifndef OPENSSL_NO_DH
-	    !ENGINE_set_DH(e, DH_get_default_method()) ||
+/* TODO add ECDH
+			!ENGINE_set_ECDH(e, PKCS11_get_ecdh_method()) ||
+*/
 #endif
-	    !ENGINE_set_RAND(e, RAND_SSLeay()) ||
-#if 0
-	    !ENGINE_set_BN_mod_exp(e, BN_mod_exp) ||
-#endif
-	    !ENGINE_set_load_pubkey_function(e, pkcs11_load_public_key) ||
-	    !ENGINE_set_load_privkey_function(e, pkcs11_load_private_key)) {
+			!ENGINE_set_load_pubkey_function(e, pkcs11_load_public_key) ||
+			!ENGINE_set_load_privkey_function(e, pkcs11_load_private_key)) {
 		return 0;
 	} else {
 		return 1;
@@ -213,5 +197,7 @@ static int bind_fn(ENGINE * e, const char *id)
 	return 1;
 }
 
-IMPLEMENT_DYNAMIC_CHECK_FN();
-IMPLEMENT_DYNAMIC_BIND_FN(bind_fn);
+IMPLEMENT_DYNAMIC_CHECK_FN()
+IMPLEMENT_DYNAMIC_BIND_FN(bind_fn)
+
+/* vim: set noexpandtab: */
